/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "steghide.h"

/* arguments */
int args_action = 0 ;
int args_space_sthdr = -1 ;
int args_encryption_sthdr = ENC_NONE ;
char *args_passphrase_sthdr = NULL ;
int args_encryption_data = ENC_NONE ;
char *args_passphrase_data = NULL ;
char *args_fn_cnt ;
char *args_fn_stg ;
char *args_fn_pln ;

/* indicates which arguments are already parsed */
int arggiven_action = 0 ;
int arggiven_method = 0 ;

int enc_sthdrfordata = 0 ;

/* static int getline (void) ; */
static void parsearguments (int argc, char *argv[]) ;
static void printhelp (void) ;
static void writedata (const char *cntfilename, const char *stgfilename, const char *plnfilename) ;
static void readdata (const char *stgfilename, const char *plnfilename) ;
static int getline(void) ;

int main (int argc, char *argv[])
{
	int iactive = 0 ;

	sthdr.stgmask = 1 ;
	
	if ((argc == 1) || (strncmp (argv[1], "-i", 2) == 0) || (strncmp (argv[1], "--interactive", 13) == 0)) {
		printf ("steghide 0.3-1\n") ;
		printf ("Copyright 1999, Stefan Hetzl\n") ;
		printf ("steghide is covered by the GNU General Public License\n") ;
		while (getline())
			;

		iactive = 1 ;
	}
	else
		parsearguments (argc, argv) ;

	if ((args_space_sthdr == -1) && (args_action == ACTN_EMBED)) {
		srand ((unsigned int) time (NULL)) ;
		args_space_sthdr = (int) (8.0 * rand() / (RAND_MAX + 1.0)) ;
	}

	if (enc_sthdrfordata) {
		args_passphrase_data = args_passphrase_sthdr ;
		args_encryption_data = args_encryption_sthdr ;
	}

	if (arggiven_action == 0)
		perr (ERR_WRONGARG) ;

	if (args_fn_cnt == NULL)
		if (args_action == ACTN_EMBED) {
			if ((args_fn_cnt = malloc (2)) == NULL)
				perr (ERR_MEMALLOC) ;
			strcpy (args_fn_cnt, "-") ;
		}

	if (args_fn_stg == NULL) {
		if ((args_fn_stg = malloc (2)) == NULL)
			perr (ERR_MEMALLOC) ;
		strcpy (args_fn_stg, "-") ;
	}

	if (args_fn_pln == NULL) {
		if ((args_fn_pln = malloc (2)) == NULL)
			perr (ERR_MEMALLOC) ;
		strcpy (args_fn_pln, "-") ;
	}

	if (args_action == ACTN_EMBED) {
		if (arggiven_method == 0)
			perr (ERR_WRONGARG) ;
		if ((strcmp (args_fn_cnt, "-") == 0) && (strcmp (args_fn_pln, "-") == 0))
			perr (ERR_WRONGARG) ;
	}

	switch (args_action) {
		case ACTN_EMBED:
		writedata (args_fn_cnt, args_fn_stg, args_fn_pln) ;
		break ;

		case ACTN_EXTRACT:
		readdata (args_fn_stg, args_fn_pln) ;
		break ;
	}

	if (args_fn_cnt != NULL)
		free (args_fn_cnt) ;
	free (args_fn_pln) ;
	free (args_fn_stg) ;
	if (args_passphrase_sthdr != NULL)
		free (args_passphrase_sthdr) ;
	if (!enc_sthdrfordata)
		if (args_passphrase_data != NULL)
			free (args_passphrase_data) ;

	if (iactive)
		printf ("done.\n") ;

	return 0 ;
}

static int getline(void)
{
	char line[512] ;
	char *argv[20] ;
	int argc, i = 0, arg = 1, argpos = 0 ;
	
	printf ("(%s) ", PROGNAME) ;

	fgets (line, 512, stdin) ;

	if ((strncmp (line, "start", 5) == 0) || (strncmp (line, "ok", 2) == 0))
		return 0 ;

	if ((strncmp (line, "quit", 4) == 0) || (strncmp (line, "cancel", 6) == 0))
		exit (EXIT_SUCCESS) ;

	if (strncmp (line, "help", 4) == 0) {
		printhelp() ;
		return 1 ;
	}

	while (line[i] != '\n') {
		if ((argv[arg] = malloc (512)) == NULL)
			perr (ERR_MEMALLOC) ;

		while ((line[i] != ' ') && (line[i] != '\t') && (line[i] != '\n') && (line[i] != '\"'))
			argv[arg][argpos++] = line[i++] ;
		argv[arg][argpos] = '\0' ;

		if ((line[i] == ' ') || (line[i] == '\t')) {
			while ((line[i] == ' ') || (line[i] == '\t'))
				i++ ;
		}

		else if (line[i] == '\"') {
			i++ ;
			while (line[i] != '\"')
				argv[arg][argpos++] = line[i++] ;
			argv[arg][argpos] = '\0' ;

			i++ ;
			while ((line[i] == ' ') || (line[i] == '\t'))
				i++ ;
		}

		arg++ ;
		argpos = 0 ;
	}

	argc = arg ;
	parsearguments (argc, argv) ;

	for (i = 1 ; i < argc ; i++)
		free (argv[i]) ;

	return 1 ;
}

/* parses command line arguments */
static void parsearguments (int argc, char* argv[])
{
	int i, j ;

	for (i = 1; i < argc; i++) {
		if ((strncmp (argv[i], "-w\0", 3) == 0) || (strncmp (argv[i], "--write\0", 8) == 0) || (strncmp (argv[i], "--embed\0", 8) == 0)) {
			args_action = ACTN_EMBED ;
			if (arggiven_action)
				perr (ERR_WRONGARG) ;
			else
				arggiven_action = 1 ;
		}

		else if ((strncmp (argv[i], "-r\0", 3) == 0) || (strncmp (argv[i], "--read\0", 7) == 0) || (strncmp (argv[i], "--extract\0", 10) == 0)) {
			args_action = ACTN_EXTRACT ;
			if (arggiven_action)
				perr (ERR_WRONGARG) ;
			else
				arggiven_action = 1 ;
		}

		else if ((strncmp (argv[i], "-s\0", 3) == 0) || (strncmp (argv[i], "--space\0", 8) == 0)) {
			i++ ;

			if (strncmp (argv[i], "cnst\0", 5) == 0) {
				unsigned char num ;

				if (arggiven_method)
					perr (ERR_WRONGARG) ;
				else
					arggiven_method = 1 ;

				sthdr.spacemtd = SPCMTD_CNST ;

				i++ ;
				num = 0 ;
				j = 0 ;
				while (isdigit (argv[i][j]))
					num = num * 10 + (argv[i][j++] - '0') ;
				sthdr.spcmtdinfo.cnst.space = num ;
			}

			else if (strncmp (argv[i], "linc\0", 5) == 0) {
				unsigned long num ;

				if (arggiven_method)
					perr (ERR_WRONGARG) ;
				else
					arggiven_method = 1 ;

				sthdr.spacemtd = SPCMTD_LINC ;

				i++ ;
				num = 0 ;
				j = 0 ;
				while (isdigit (argv[i][j]))
					num = num * 10 + (argv[i][j++] - '0') ;
				sthdr.spcmtdinfo.linc.seed = num ;

				i++ ;
				num = 0 ;
				j = 0 ;
				while (isdigit (argv[i][j]))
					num = num * 10 + (argv[i][j++] - '0') ;
				if (num > 255)
					sthdr.spcmtdinfo.linc.maxspace = 255 ;
				else
					sthdr.spcmtdinfo.linc.maxspace = (unsigned char) num ;
			}

			else
				perr (ERR_WRONGARG) ;
		}

		else if ((strncmp (argv[i], "-e\0", 3) == 0) || (strncmp (argv[i], "--encryption\0", 13) == 0)) {
			i++ ;
			if (strncmp (argv[i], "blowfish\0", 9) == 0) {
				args_encryption_data = ENC_BLOWFISH ;
				i++ ;
				if ((args_passphrase_data = malloc (strlen (argv[i]) + 1)) == NULL)
					perr (ERR_MEMALLOC) ;
				strcpy (args_passphrase_data, argv[i]) ;
				j = 0 ;
				while (argv[i][j] != '\0')
					argv[i][j++] = ' ' ;
			}
			else if (strncmp (argv[i], "sthdr\0", 6) == 0) {
				enc_sthdrfordata = 1 ;
			}
			else if (strncmp (argv[i], "none\0", 5) == 0) {
				args_encryption_data = ENC_NONE ;
			}
			else {
				perr (ERR_WRONGARG) ;
			}
		}

		else if ((strncmp (argv[i], "-H\0", 3) == 0) || (strncmp (argv[i], "--stegoheader\0", 14) == 0)) {
			unsigned char num ;

			i++ ;
			if (isdigit (*argv[i])) {
				num = 0 ;
				j = 0 ;

				while (isdigit (argv[i][j]))
					num = num * 10 + (argv[i][j++] - '0') ;

				if (num > 7)
					args_space_sthdr = 7 ;
				else
					args_space_sthdr = num ;

				i++ ;
			}
			else {
				if (args_action == ACTN_EMBED) {
					srand ((unsigned int) time (NULL)) ;
					args_space_sthdr = (int) (8.0 * rand() / (RAND_MAX + 1.0)) ;
				}
			}
			
			if (strncmp (argv[i], "blowfish\0", 9) == 0) {
				args_encryption_sthdr = ENC_BLOWFISH ;
				i++ ;
				if ((args_passphrase_sthdr = malloc (strlen (argv[i]) + 1)) == NULL)
					perr (ERR_MEMALLOC) ;
				strcpy (args_passphrase_sthdr, argv[i]) ;
				j = 0 ;
				while (argv[i][j] != '\0')
					argv[i][j++] = ' ' ;
					
			}
			else if (strncmp (argv[i], "none\0", 5) == 0) {
				args_encryption_sthdr = ENC_NONE ;
			}
			else
				perr (ERR_WRONGARG) ;
		}

		else if ((strncmp (argv[i], "-m\0", 3) == 0) || (strncmp (argv[i], "--mask\0", 7) == 0)) {
			unsigned char num ;

			i++ ;
			num = 0 ;
			j = 0 ;
			while (isdigit (argv[i][j]))
				num = num * 10 + (argv[i][j++] - '0') ;
			sthdr.stgmask = num & 0xFF ;
		}

		else if ((strncmp (argv[i], "-cf\0", 4) == 0) || (strncmp (argv[i], "--containerfile\0", 16) == 0)) {
			i++ ;
			if ((args_fn_cnt = malloc (strlen (argv[i]) + 1)) == NULL)
				perr (ERR_MEMALLOC) ;
			strcpy (args_fn_cnt, argv[i]) ;
		}

		else if ((strncmp (argv[i], "-sf\0", 4) == 0) || (strncmp (argv[i], "--stegofile\0", 12) == 0)) {
			i++ ;
			if ((args_fn_stg = malloc (strlen (argv[i]) + 1)) == NULL)
				perr (ERR_MEMALLOC) ;
			strcpy (args_fn_stg, argv[i]) ;
		}

		else if ((strncmp (argv[i], "-pf\0", 4) == 0) || (strncmp (argv[i], "--plainfile\0", 12) == 0)) {
			i++ ;
			if ((args_fn_pln = malloc (strlen (argv[i]) + 1)) == NULL)
				perr (ERR_MEMALLOC) ;
			strcpy (args_fn_pln, argv[i]) ;
		}

		else if ((strncmp (argv[i], "-v\0", 3) == 0) || (strncmp (argv[i], "--version\0", 10) == 0)) {
			printf ("steghide version 0.3-1\n") ;
			exit (EXIT_SUCCESS) ;
		}

		else if ((strncmp (argv[i], "-h\0", 3) == 0) || (strncmp (argv[i], "--help\0", 7) == 0)) {
			printf ("steghide version 0.3-1\n\n") ;
			printf (" -i, --interactive      enter interactive mode\n") ;
			printhelp () ;
			exit (EXIT_SUCCESS) ;
		}

		else {
			perr (ERR_WRONGARG) ;
		}
	}
}

static void printhelp ()
{
	printf (" -w, --write, --embed   embed plain data in container data\n") ;
	printf (" -r, --read, --extract  extract plain data from stego data\n\n") ;
	printf (" -s, --space            select method to calculate space between two stegobytes\n") ;
	printf ("                        -s cnst <a> use constant space <a>\n") ;
	printf ("                        -s linc <a> <b> use pseudo random spaces with seed <a>\n") ;
	printf ("                        and maximum value <b>\n") ;
	printf (" -e, --encryption       use encryption on plainfile\n") ;
	printf ("                        -e blowfish \"<pass>\" use the blowfish algorithm with\n") ;
	printf ("                        <pass> (<pass> is hashed to key using MD5.)\n") ;
	printf ("                        -e sthdr use encryption defined with -H\n") ;
	printf (" -H, --stegoheader      parameters for the stego header\n") ;
	printf ("                        -H [<space>] blowfish \"<pass>\" use space <space> to\n") ;
	printf ("                        embed stego header, use blowfish to encrypt stegoheader\n") ;
	printf ("                        with <pass>. If <space> is omitted, a random number\n") ;
	printf ("                        between 0 and 7 is chosen\n") ;
	printf (" -m, --mask             -m <a> use <a> as mask to hide bits in container data\n\n") ;
	printf (" -cf, --containerfile   -cf <f> use <f> as container file\n") ;
	printf (" -sf, --stegofile       -sf <f> use <f> as stego file\n") ;
	printf (" -pf, --plainfile       -pf <f> use <f> as plain file\n\n") ;
	printf (" -v, --version          print version information\n") ;
	printf (" -h, --help             print this help\n\n") ;	
	printf ("-w and -s must be given to embed data.\n") ;
	printf ("-r must be given to extract data.\n") ;
	printf ("If no file or filename \"-\" is given, standard input or standard output is used.\n") ;
	printf ("If data is embedded, at least one of -cf and -pf must be a real file.\n") ;
}

/* calls functions to embed plain data in container data and save as stego data */
static void writedata (const char *cntfilename, const char *stgfilename, const char *plnfilename)
{
	CNTFILE *cntfile, *stgfile ;
	PLNFILE *plnfile ;

	cntfile = readcntfile (cntfilename) ;

	plnfile = readplnfile (plnfilename) ;

	sthdr.nbytesplain = bufgetdatalen (plnfile->plnbuflhead) ;

	encrypt_data (plnfile, args_encryption_data, args_passphrase_data) ;

	writesthdr (cntfile->cntbuflhead, args_encryption_sthdr, args_passphrase_sthdr) ;

	embeddata (cntfile->cntbuflhead, plnfile->plnbuflhead) ;

	stgfile = createstgfile (cntfile, stgfilename) ;

	writecntfile (stgfile) ;

	cleanupcntfile (cntfile, FSS_NO) ;
	cleanupcntfile (stgfile, FSS_YES) ;
	cleanupplnfile (plnfile) ;

	return ;
}

/* calls functions to extract (and save) plain data from stego data */
static void readdata (const char *stgfilename, const char *plnfilename)
{
	CNTFILE *stgfile ;
	PLNFILE *plnfile ;

	stgfile = readcntfile (stgfilename) ;

	readsthdr (stgfile->cntbuflhead, args_encryption_sthdr, args_passphrase_sthdr) ;

	plnfile = createplnfile (plnfilename) ;

	plnfile->plnbuflhead = extractdata (stgfile->cntbuflhead) ;

	decrypt_data (plnfile, args_encryption_data, args_passphrase_data) ;

	writeplnfile (plnfile) ;

	cleanupcntfile (stgfile, FSS_YES) ;
	cleanupplnfile (plnfile) ;

	return ;
}

/* prints error message on stderr and exits with error code */
void perr (int errn)
{
	char errmsg[256] = "" ;

	switch (errn) {
		case ERR_OK:
		strcpy (errmsg, "perr() has been called, but no error occured") ;
		break ;

		case ERR_WRONGARG:
		strcpy (errmsg, "wrong argument syntax - type \"steghide -h\" for help") ;
		break ;

		case ERR_CNTOPEN:
		strcpy (errmsg, "can not open container file") ;
		break ;

		case ERR_STGOPEN:
		strcpy (errmsg, "can not open stego file") ;
		break ;

		case ERR_PLNOPEN:
		strcpy (errmsg, "can not open plain file") ;
		break ;

		case ERR_CNTREAD:
		strcpy (errmsg, "can not read from container file") ;
		break ;

		case ERR_STGREAD:
		strcpy (errmsg, "can not read from stego file") ;
		break ;

		case ERR_PLNREAD:
		strcpy (errmsg, "can not read from plain file") ;
		break ;

		case ERR_STGWRITE:
		strcpy (errmsg, "can not write to stego file") ;
		break ;

		case ERR_PLNWRITE:
		strcpy (errmsg, "can not write to plain file") ;
		break ;

		case ERR_CNTCLOSE:
		strcpy (errmsg, "can not close container file") ;
		break ;

		case ERR_STGCLOSE:
		strcpy (errmsg, "can not close stego file") ;
		break ;

		case ERR_PLNCLOSE:
		strcpy (errmsg, "can not close plain file") ;
		break ;

		case ERR_CNTUNKNOWNFFORMAT:
		strcpy (errmsg, "the container file has an unknown file format") ;
		break ;

		case ERR_STGUNKNOWNFFORMAT:
		strcpy (errmsg, "the stego file has an unknown file format") ;
		break ;

		case ERR_CNTUNSUPBMPFILE:
		strcpy (errmsg, "the container file is an unsupported bmp file") ;
		break ;

		case ERR_STGUNSUPBMPFILE:
		strcpy (errmsg, "the stego file is an unsupported bmp file") ;
		break ;

		case ERR_CNTBMPWITHCOMPR:
		strcpy (errmsg, "the container file is a bmp file that uses compression") ;
		break ;

		case ERR_STGBMPWITHCOMPR:
		strcpy (errmsg, "the stego file is a bmp file that uses compression") ;
		break ;

		case ERR_CNTUNSUPWAVFILE:
		strcpy (errmsg, "the container file is an unsupported wav file") ;
		break ;

		case ERR_STGUNSUPWAVFILE:
		strcpy (errmsg, "the stego file is an unsupported wav file") ;
		break ;

		case ERR_CNTBPSNOTMULTOF8:
		strcpy (errmsg, "the container file is a wav file with a bits per sample rate that is not a multiple of eight") ;
		break ;

		case ERR_STGBPSNOTMULTOF8:
		strcpy (errmsg, "the stego file is a wav file with a bits per sample rate that is not a multiple of eight") ;
		break ;

		case ERR_MEMALLOC:
		strcpy (errmsg, "can not allocate memory") ;
		break ;

		case ERR_CNTTOOSHORT:
		strcpy (errmsg, "the container file is too short to embed plain data") ;
		break ;

		case ERR_STGTOOSHORT:
		strcpy (errmsg, "the stego file is too short to contain plain data of specified length") ;
		break ;

		default:
		strcpy (errmsg, "unexpected error! this should not happen") ;
		break ;
	}
	fprintf (stderr, "%s: %s\n", PROGNAME, errmsg) ;
	fprintf (stderr, "%s: errorcode %d, quitting.\n", PROGNAME, errn) ;
	exit (errn) ;
}
