/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "steghide.h"

STEGOHEADER sthdr = { '\0', { { '\0' } }, '\0', 0 } ;

static int setbits (int cntbyte, int plnbits) ;
static int getbits (int stgbyte) ;
static int nstgbits (void) ;
static int spacetonextbyte (int action) ;
static int space_cnst (int action) ;
static int space_linc (int action) ;
static unsigned long mult (unsigned long p, unsigned long q) ;

/* embed plain data in container data (resulting in stego data) */
void embeddata (BUFFER *cntbuflhead, BUFFER *plnbuflhead)
{
	int cntbyte, stgbyte, plnbits ;
	int bit, i ;
	long bitsread = 0 ;

	spacetonextbyte (SP_RESET) ;

	while (((bitsread / 8) < bufgetdatalen (plnbuflhead)) && (!bufendofbuf (cntbuflhead))) {
		plnbits = 0 ;
		for (i = 0 ; i < nstgbits() ; i++)
			if ((bit = bufgetbit (plnbuflhead)) != ENDOFBUF) {
				plnbits |= (bit << i) ;
				bitsread++ ;
			}

		cntbyte = bufgetbyte (cntbuflhead) ;
		stgbyte = setbits (cntbyte, plnbits) ;
		bufseek (cntbuflhead, -1, BUFSEEK_CUR) ;
		bufsetbyte (stgbyte, cntbuflhead) ;

		if (bufseek (cntbuflhead, spacetonextbyte (SP_GETVALUE), BUFSEEK_CUR) == ENDOFBUF)
			perr (ERR_CNTTOOSHORT) ;
	}

	bufseek (cntbuflhead, 0, BUFSEEK_SET) ;
	bufseek (plnbuflhead, 0, BUFSEEK_SET) ;

	return ;
}

/* extracts plain data (return value) from stego data */
BUFFER *extractdata (BUFFER *stgbuflhead)
{
	BUFFER *plnbuflhead ;
	int plnbits, i ;
	long bitsread = 0, len ;

	spacetonextbyte (SP_RESET) ;
	plnbuflhead = initbuflist() ;

	if (args_encryption_data == ENC_BLOWFISH) {
		if (sthdr.nbytesplain % 8 != 0)
			len = ((sthdr.nbytesplain / 8) + 1) * 8 ;
	}
	else
		len = sthdr.nbytesplain ;

	while ((bitsread / 8) < len) {
		plnbits = getbits (bufgetbyte (stgbuflhead)) ;
		for (i = 0 ; i < nstgbits() ; i++) {
			bufsetbit (((plnbits & (1 << i)) >> i), plnbuflhead) ;
			bitsread++ ;
		}

		if (bufseek (stgbuflhead, spacetonextbyte (SP_GETVALUE), BUFSEEK_CUR) == ENDOFBUF)
			perr (ERR_STGTOOSHORT) ;
	}
			
	bufseek (stgbuflhead, 0, BUFSEEK_SET) ;
	bufseek (plnbuflhead, 0, BUFSEEK_SET) ;

	return plnbuflhead ;
}

/* embeds the stegoheader in the container data */
void writesthdr (BUFFER *cntbuflhead, int encryption, char *passphrase)
{
	unsigned char hdrbuf [STHDR_BLOCKS_BLOWFISH * 8] ;
	int datalen, bit, byte = 0 ;

	hdrbuf[byte++] = sthdr.spacemtd ;

	switch (sthdr.spacemtd) {
		case SPCMTD_CNST:
		hdrbuf[byte++] = sthdr.spcmtdinfo.cnst.space ;
		break ;

		case SPCMTD_LINC:
		cp32ul2uc_be (&hdrbuf[byte], sthdr.spcmtdinfo.linc.seed) ;
		byte += 4 ;
		hdrbuf[byte++] = sthdr.spcmtdinfo.linc.maxspace ;
		break ;
	}

	hdrbuf[byte++] = sthdr.stgmask ;
	cp32ul2uc_be (&hdrbuf[byte], sthdr.nbytesplain) ;
	byte += 4 ;

	srand ((unsigned int) time (NULL)) ;
	while (byte < STHDR_BLOCKS_BLOWFISH * 8)
		hdrbuf[byte++] = (int) (256.0 * rand() / (RAND_MAX + 1.0)) ;

	encrypt_sthdr (hdrbuf, byte, encryption, passphrase) ;

	datalen = byte ;
	bufsetbitpos (0, cntbuflhead) ;

	for (bit = 0 ; bit < 3 ; bit++) {
		bufsetbit (((args_space_sthdr & (1 << bit)) >> bit), cntbuflhead) ;
		bufseek (cntbuflhead, 1, BUFSEEK_CUR) ;
		bufsetbitpos (0, cntbuflhead) ;
	}

	for (byte = 0 ; byte < datalen ; byte++)
		for (bit = 0 ; bit < 8 ; bit++) {
			bufsetbit (((hdrbuf[byte] & (1 << bit)) >> bit), cntbuflhead) ;
			bufseek (cntbuflhead, args_space_sthdr + 1, BUFSEEK_CUR) ;
			bufsetbitpos (0, cntbuflhead) ;
		}

	return ;
}

/* extracts the stegoheader from the stego data */
void readsthdr (BUFFER *stgbuflhead, int encryption, char *passphrase)
{
	unsigned char hdrbuf [STHDR_BLOCKS_BLOWFISH * 8] ;
	int bit, byte ;

	args_space_sthdr = 0 ;
	for (bit = 0 ; bit < 3 ; bit++) {
		args_space_sthdr |= bufgetbit (stgbuflhead) << bit ;
		bufseek (stgbuflhead, 1, BUFSEEK_CUR) ;
		bufsetbitpos (0, stgbuflhead) ;
	}

	for (byte = 0 ; byte < STHDR_BLOCKS_BLOWFISH * 8 ; byte++) {
		hdrbuf[byte] = 0 ;
		for (bit = 0; bit < 8; bit++) {
			hdrbuf[byte] |= bufgetbit (stgbuflhead) << bit ;
			bufseek (stgbuflhead, args_space_sthdr + 1, BUFSEEK_CUR) ;
			bufsetbitpos (0, stgbuflhead) ;
		}
	}

	decrypt_sthdr (hdrbuf, encryption, passphrase) ;

	byte = 0 ;

	sthdr.spacemtd = hdrbuf[byte++] ;

	switch (sthdr.spacemtd) {
		case SPCMTD_CNST:
		sthdr.spcmtdinfo.cnst.space = hdrbuf[byte++] ;
		break ;

		case SPCMTD_LINC:
		cp32uc2ul_be (&sthdr.spcmtdinfo.linc.seed, &hdrbuf[byte]) ;
		byte += 4 ;
		sthdr.spcmtdinfo.linc.maxspace = hdrbuf[byte++] ;
		break ;
	}

	sthdr.stgmask = hdrbuf[byte++] ;
	cp32uc2ul_be (&sthdr.nbytesplain, &hdrbuf[byte]) ;
	/*
	byte += 4 ;
	*/

	return ;
}

/* writes plnbits to cntbyte as determined by sthdr.stgmask resulting in return value stgbyte */
static int setbits (int cntbyte, int plnbits)
{
	int maskbitpos = 0, plnbitpos = 0 ;
	int stgbyte, bit ;

	stgbyte = cntbyte & (255 - sthdr.stgmask) ;

	while (maskbitpos < 8) {
		while ((!(sthdr.stgmask & (1 << maskbitpos))) && (maskbitpos < 8))
			maskbitpos++ ;

		if (maskbitpos == 8)
			break ;

		bit = (plnbits & (1 << plnbitpos)) >> plnbitpos ;
		stgbyte |= bit << maskbitpos ;
		plnbitpos++ ;
		maskbitpos++ ;
	}

	return stgbyte ;
}

/* reads plain bits (return value) from stgbyte as determined by sthdr.stgmask */
static int getbits (int stgbyte)
{
	int maskbitpos = 0, plnbitpos = 0 ;
	int plnbits = 0, bit ;

	while (maskbitpos < 8) {
		while ((!(sthdr.stgmask & (1 << maskbitpos))) && (maskbitpos < 8))
			maskbitpos++ ;

		if (maskbitpos == 8)
			break ;

		bit = (stgbyte & (1 << maskbitpos)) >> maskbitpos ;
		plnbits |= bit << plnbitpos ;
		plnbitpos++ ;
		maskbitpos++ ;
	}

	return plnbits ;
}

/* returns the number of set bits in sthdr.stgmask */
static int nstgbits (void)
{
	int i, n = 0 ;

	for (i = 0; i < 8; i++)
		if (sthdr.stgmask & (1 << i))
			n++ ;

	return n ;
}

/* determines space betweeen two stego bytes */
static int spacetonextbyte (int action)
{
	int retval ;

	if (action == SP_RESET) {
		space_linc (SP_RESET) ;
		retval = ERR_OK ;
	}

	if (action == SP_GETVALUE) {
		switch (sthdr.spacemtd) {
			case SPCMTD_CNST:
			retval = space_cnst (SP_GETVALUE) ;
			break ;

			case SPCMTD_LINC:
			retval = space_linc (SP_GETVALUE) ;
			break ;
		}
	}

	return retval ;
}

/* returns space between two stegobytes using the CNST spacemtd */
static int space_cnst (int action)
{
	int retval ;

	if (action == SP_RESET)
		retval = ERR_OK ;

	if (action == SP_GETVALUE)
		retval = sthdr.spcmtdinfo.cnst.space ;

	return retval ;
}

/* returns space between two stegobytes using the LINC spacemtd */
static int space_linc (int action)
{
	static unsigned long a ;
	static int maxspace ;
	int retval ;

	if (action == SP_RESET) {
		a = sthdr.spcmtdinfo.linc.seed ;
		maxspace = sthdr.spcmtdinfo.linc.maxspace ;
		retval = ERR_OK ;
	}

	if (action == SP_GETVALUE) {
		a = (mult (a, LINC_b) + 1) % LINC_m ;
		retval = (int) (((double) a / LINC_m) * maxspace) ;
	}

	return retval ;
}

/* multiply p and q avoiding an overflow */
static unsigned long mult (unsigned long p, unsigned long q)
{
	unsigned long p1, p0, q1, q0 ;

	p1 = p / LINC_m1 ;
	p0 = p % LINC_m1 ;
	q1 = q / LINC_m1 ;
	q0 = q % LINC_m1 ;

	return (((p0 * q1 + p1 + q0) % LINC_m1) * LINC_m1 + p0 * q0) % LINC_m ;
}
